import time, random, math, os

import pygame
from pygame.locals import *

from ai import *

# various important things
sounds = {}
for file in os.listdir('data'):
    if not file.endswith('.ogg'): continue
    name = os.path.splitext(file)[0]
    sounds[name] = pygame.mixer.Sound(os.path.join('data', file))

modenine = pygame.font.Font('data/MODENINE.TTF', 12)
modenine_small = pygame.font.Font('data/MODENINE.TTF', 8)

# cache rendered numbers
text_nums = {}

# codes from the code tiles
CODE_GO = 1
CODE_LIT = 2
CODE_UNLIT = 3
CODE_GHOST = 4
CODE_DEATH = 17

# tiles to mod the map with
TILE_LIGHT_LIT = 16
TILE_LIGHT_UNLIT = 19

black = (0,0,0)
white = (255,255,255)


WALK_SOUND_EVENT = USEREVENT + 1
REPAIR_SOUND_EVENT = USEREVENT + 2

class Quit(Exception):
    pass


class Actor:
    saved_screen_rect = None
    def __init__(self, position):
        self.position = position
        self.move = [0, 0]

    def getTilePos(self):
        x, y = self.position
        return int(x), int(y)

    def getScreenPos(self):
        x, y = self.position
        return int(x*16), int(y*16)

    def getRect(self):
        x, y = self.position
        return Rect((int(x*16), int(y*16), 16, 16))

    def testCollision(self, other):
        if not isinstance(other, pygame.Rect): other = other.getRect()
        return self.getRect().colliderect(other)

    def repair(self, game):
        # repair screen in current screen pos
        if not game.redrawn and self.saved_screen_rect is not None:
            game.screen.blit(game.drawn_level, self.saved_screen_rect[:2],
                self.saved_screen_rect)
            game.screen.blit(game.shadow_map, self.saved_screen_rect[:2],
                self.saved_screen_rect)

    def render(self, game):
        # draw actor to new screen pos
        screen_pos = self.getScreenPos()
        screen_rect = screen_pos + (16, 16)

        # save off the repair info
        self.saved_screen_rect = screen_rect

        # and finally, draw
        game.screen.blit(game.tiles, screen_pos, self.getTile(game))
        game.screen.blit(game.shadow_map, screen_pos, screen_rect)

class Player(Actor):
    saved_shadow_map = None
    saved_shadow_map_pos = None

    walk_channel = None
    repair_channel = None

    SPEED = .25

    move_left = move_right = move_up = move_down = False
    def startMoveLeft(self): self.move_left = True
    def startMoveRight(self): self.move_right = True
    def startMoveUp(self): self.move_up = True
    def startMoveDown(self): self.move_down = True
    def stopMoveLeft(self): self.move_left = False
    def stopMoveRight(self): self.move_right = False
    def stopMoveUp(self): self.move_up = False
    def stopMoveDown(self): self.move_down = False

    def animate(self, game):
        px, py = self.position
        dx = dy = 0
        if self.move_left: dx -= self.SPEED
        if self.move_right: dx += self.SPEED
        if self.move_up: dy -= self.SPEED
        if self.move_down: dy += self.SPEED
        if not (dx or dy):
            return False

        # death-of-a-thousand-rects
        # compare our intended destination with surrounding tiles and if
        # any collide, we don't move
        ex, ey = px+dx, py+dy
        dest = Rect((px+dx) *16, (py+dy) *16, 16, 16)
        tx, ty = int(px), int(py)
        for i in range(-1, 2):
            for j in range(-1, 2):
                tile, back, code = game.level[tx+i][ty+j]
                if not tile and not code: continue

                check = Rect((tx+i) *16, (ty+j) *16, 16, 16)
                if not dest.colliderect(check): continue

                if code == CODE_DEATH:
                     return True

                if not tile: continue

                if i: dx = 0
                if j: dy = 0
                if tile == TILE_LIGHT_UNLIT:
                    if self.repair_channel is None:
                        sound = random.choice(('hammer', 'drill',
                            'socket-wrench'))
                        self.repair_channel = sounds[sound].play()
                        if self.repair_channel is not None:
                            self.repair_channel.set_endevent(REPAIR_SOUND_EVENT)
                    game.progress.fixLight(game.lights[(tx+i, ty+j)], self)

        if dx or dy:
            #if self.walk_channel is None:
            #    self.walk_channel = game.sounds['walk'].play()
            #    self.walk_channel.set_endevent(WALK_SOUND_EVENT)

            nx, ny = px + dx, py + dy
            self.position = (nx, ny)
        #else:
            #if self.walk_channel is not None:
            #    self.walk_channel.fadeout(100)
            #    self.walk_channel = None

        return False

    def getTile(self, game):
        return (4*16, 0, 16, 16)

    def modLightMap(self, game):
        # repair light map in current screen pos
        if not game.redrawn and self.saved_shadow_map_pos:
            game.shadow_map.blit(self.saved_shadow_map,
                self.saved_shadow_map_pos)
            game.screen.blit(self.saved_shadow_map, self.saved_shadow_map_pos)

        # draw actor to new screen pos
        x, y = self.getScreenPos()
        light_pos = (x-24, y-24)
        light_rect = light_pos + (64, 64)

        # save off the repair info
        self.saved_shadow_map_pos = light_pos
        if self.saved_shadow_map is None:
            self.saved_shadow_map = pygame.Surface((64, 64), HWSURFACE,
                game.shadow_map)
        self.saved_shadow_map.fill(white)
        self.saved_shadow_map.blit(game.shadow_map, (0, 0), light_rect)
        self.saved_shadow_map.set_colorkey(white)

        # fresh draw
        game.screen.blit(game.drawn_level, light_pos, light_rect)

        # mod the light map
        pygame.draw.circle(game.shadow_map, white, (x + 8, y + 8), 32, 0)
        game.shadow_map.set_colorkey(white)

        # draw the new shadow map
        game.screen.blit(game.shadow_map, light_pos, light_rect)

class Ghost(Actor):
    LOOK_LEFT = 0
    LOOK_UP = 1
    LOOK_RIGHT = 2
    LOOK_DOWN = 3

    MIN_LT_DIST = 8
    MIN_SHOOT_DIST = 10
    RECHARGE_TIME = 1      # secs for ghost to recharge

    looking = LOOK_LEFT
    zapping = None
    recharging = 0

    came_from_tile = None
    going_to_tile = None

    SPEED = .1

    saved_screen = None

    def animate(self, game):
        self.zapping = None

        if self.going_to_tile:
            tx, ty = self.going_to_tile
            x, y = self.position
            dx, dy = tx-x, ty-y

            # figure our desired move distance and direction
            dist = math.sqrt(dx**2 + dy**2)
            angle = math.atan2(dx, dy)

            # figure how far we're actually going to move
            move = min(dist, self.SPEED)
            x = x + math.sin(angle) * move
            y = y + math.cos(angle) * move

            # have we hit the destination?
            if abs(tx-x) < 0.01 and abs(ty-y) < 0.01:
                self.position = self.going_to_tile
                self.going_to_tile = None
            else:
                self.position = (x, y)
                # moving, can't do anything else
                return

        # ok, are we running away from any lights?
        ldist = {}
        for tile_pos, light in game.lights.items():
            tx, ty = tile_pos
            if not light: continue    # light not on
            dist = calculateDistance(self.position, tile_pos)
            if dist < self.MIN_LT_DIST:
                try:
                    self.going_to_tile = findWorse(game.lightff,
                        self.getTilePos(), self.came_from_tile)
                except DestroyGhost:
                    game.ghosts.remove(self)
                return
            ldist[dist] = light

        if self.recharging:
            if self.recharging < time.time():
                self.recharging = None
            # can't do anything else
            return

        # find something to zap
        l = ldist.items()
        l.sort()
        for dist, light in l:
            if light and dist < self.MIN_SHOOT_DIST:
                self.zapping = light.getTilePos()
                break
        else:
            tpos = self.getTilePos()
            self.came_from_tile = tpos
            try:
                self.going_to_tile = findBetter(game.lightff, tpos,
                    self.came_from_tile)
            except DestroyGhost:
                game.ghosts.remove(self)

    def getTile(self, game):
#        if self.recharging:
#            return (2*16, 7*16, 16, 16)
        return (0, 7*16, 16, 16)

    def getEyesTile(self, player):
        x, y = self.position
        px, py = player.position
        ax = abs(x-px)
        ay = abs(y-py)
        if px <= x and ax >= ay: self.looking = self.LOOK_LEFT
        elif px >= x and ax >= ay: self.looking = self.LOOK_RIGHT
        elif py < y and ax < ay: self.looking = self.LOOK_UP
        elif py > y and ax < ay: self.looking = self.LOOK_DOWN
        return (self.looking*16, 6*16, 16, 16)

    def render(self, game):
        Actor.render(self, game)

        # render the eyes too
        game.screen.blit(game.tiles, self.getScreenPos(),
            self.getEyesTile(game.player))


class Blob(Actor):
    def __init__(self, position, target):
        Actor.__init__(self, position)
        self.target = target

    def getTile(self, game):
        return (3*16, 0, 16, 16)

    def animate(self, game):
        x, y = self.getTilePos()
        tx, ty = self.target
        if x > tx: x = x -1
        elif x < tx: x = x + 1
        if y > ty: y = y -1
        elif y < ty: y = y + 1
        self.position = (x, y)

        if (x,y) == self.target:
            light = game.lights[self.target]
            light.damage += random.randint(1,5)
            sounds['damage%s'%random.randint(1,3)].play()
            game.blobs.remove(self)
            game.blobs_done.append(self)
            if light.damage > 25 and random.random() > .5:
                light.ok = False
                game.level[x][y][0] = TILE_LIGHT_UNLIT
                game.drawOneLevelElement(x, y)
                return True

        return False

class Light:
    def __init__(self, tile_pos, damage=0):
        self.tile_pos = tile_pos
        self.damage = damage
        self.ok = damage == 0

    def __nonzero__(self):
        return self.ok

    def getTilePos(self):
        return self.tile_pos

    def getScreenPos(self):
        x, y = self.tile_pos
        return x*16, y*16

class ProgressDialog:
    light = None

    def __init__(self, game):
        text = modenine.render('Repairing...', 1, (100, 0, 0),
            (200, 200, 200))
        tx, ty = text.get_size()
        self.surf = pygame.Surface((tx+20, ty+50), HWSURFACE, game.screen)
        self.surf.fill((200, 200, 200))
        self.surf.blit(text, (10, 10))

        self.bwidth = tx
        self.btop = ty + 20

    def __nonzero__(self):
        return self.light is not None

    def fixLight(self, light, game):
        self.light = light

    def stopFixingLight(self):
        self.light = None

    def animate(self, game):
        self.light.damage -= 1
        if self.light.damage:
            return False

        if game.player.repair_channel:
            game.player.repair_channel.fadeout(100)
            game.player.repair_channel = None

        # light fixed
        x, y = self.light.getTilePos()
        game.level[x][y][0] = TILE_LIGHT_LIT
        game.drawOneLevelElement(x, y)
        self.light.ok = True
        self.light = None
        return True     # regen lights

    def getRect(self, game):
        rect = self.surf.get_rect()
        mx, my = self.light.getScreenPos()
        rect.center = (mx + 8, my + 8)
        return rect.clamp(game.screen.get_rect())

    covered_rect = None
    def render(self, game):
        self.surf.fill((100, 100, 100), (10, self.btop, self.bwidth, 10))
        donewidth = int(self.bwidth * (self.light.damage / 50.0))
        self.surf.fill((255, 50, 50), (10, self.btop, donewidth, 10))

        rect = self.getRect(game)
        self.covered_rect = rect
        game.screen.blit(self.surf, rect.topleft)

    def repair(self, game):
        # repair screen in current screen pos
        if not game.redrawn and self.covered_rect:
            game.screen.blit(game.drawn_level, self.covered_rect.topleft,
                self.covered_rect)
            game.screen.blit(game.shadow_map, self.covered_rect.topleft,
                self.covered_rect)

class Game:
    pause = False
    show_fps = True
    step = 0
    expose = False

    def main(self, scr, level):
        self.screen = scr

        level = pygame.image.load(level).convert_alpha(scr)
        self.level = pygame.surfarray.pixels3d(level)

        self.tiles = pygame.image.load('data/tiles.tga').convert_alpha(scr)

        self.timer = pygame.time.Clock()

        # determine game starting state
        self.ghosts = []
        self.blobs = []
        self.blobs_done = []
        self.lights = {}
        Ghost.RECHARGE_TIME = self.level[0,0,2] - 31
        for x, row in enumerate(self.level):
            for y, cell in enumerate(row):
                if cell[2] == CODE_GO:
                    self.player = Player((x, y))
                elif cell[2] == CODE_GHOST:
                    self.ghosts.append(Ghost((x,y)))
                elif cell[0] == TILE_LIGHT_UNLIT:
                    # initialise with some damage
                    self.lights[(x,y)] = Light((x,y), random.randint(20, 50))
                elif cell[0] == TILE_LIGHT_LIT:
                    self.lights[(x,y)] = Light((x,y))

        # render the level and lights to the drawn_map and shadow_map Surfaces
        self.renderLevel()
        self.renderLights()

        # draw the initial state of the screen
        self.renderPlayArea()

        # pre-render the progress dialog
        self.progress = ProgressDialog(self)

        while 1:
            val = self.timer.tick(25)
            if self.events():
                raise Quit

            if self.pause:
                continue

            # stepping
            if self.step > 1:
                continue
            if self.step:
                self.step = 2

            if self.animate():
                return False

            # check player victory
            for light in self.lights.values():
                if not light: break
            else: return True

            # check ghost victory
            for light in self.lights.values():
                if light: break
            else: return False

            self.renderMovables()

            if self.show_fps:
                self.renderFPS()
            pygame.display.flip()

    def renderPlayArea(self):
        # draw the initial state of the screen
        self.screen.blit(self.drawn_level, (0,0))
        self.screen.blit(self.shadow_map, (0,0))
        self.redrawn = True
        pygame.display.flip()

    def renderMovables(self):
        self.progress.repair(self)

        for ghost in self.ghosts:
            ghost.repair(self)
        for blob in self.blobs:
            blob.repair(self)
        for blob in self.blobs_done[:]:
            blob.repair(self)
            self.blobs_done.remove(blob)
        self.player.repair(self)

        self.player.modLightMap(self)

        for ghost in self.ghosts:
            ghost.render(self)
        for blob in self.blobs:
            blob.render(self)
        self.player.render(self)

        if self.progress:
            self.progress.render(self)

        self.redrawn = False

    def renderLevel(self):
        self.drawn_level = pygame.Surface((800, 600), HWSURFACE, self.screen)
        self.drawn_level.fill((100,100,100))
        for x, row in enumerate(self.level):
            for y, cell in enumerate(row):
                self.drawOneLevelElement(x, y)

    def drawOneLevelElement(self, x, y):
        tile, back, code = self.level[x][y]
        tiley = (back/16) * 16
        tilex = (back%16) * 16
        self.drawn_level.blit(self.tiles, (x*16, y*16), (tilex, tiley, 16, 16))
        tiley = (tile/16) * 16
        tilex = (tile%16) * 16
        self.drawn_level.blit(self.tiles, (x*16, y*16), (tilex, tiley, 16, 16))

    def renderLights(self):
        self.shadow_map = pygame.Surface((800, 576), HWSURFACE, self.screen)
        self.shadow_map.fill(black)
        width = 0
        for x, row in enumerate(self.level):
            for y, cell in enumerate(row):
                if cell[0] == TILE_LIGHT_LIT:
                    pygame.draw.circle(self.shadow_map, white,
                        (x*16+8, y*16+8), 7*16, width)
                if cell[0] == TILE_LIGHT_UNLIT:
                    self.shadow_map.blit(self.tiles, (x*16, y*16), (4*16,
                        16, 16, 16))

        self.shadow_map.set_colorkey(white)

        # now generate Ghost move flood map
        n = len(self.level[0])-1
        self.lightff = [[0]*n for row in self.level]
        for x, row in enumerate(self.level):
            for y, cell in enumerate(row):
                if cell[0] == TILE_LIGHT_LIT:
                    self.lightff[x][y] = 1
        floodFill(self.lightff, self.level)


    def renderFPS(self):
        fps = int(self.timer.get_fps())
        if fps not in text_nums:
            text = modenine.render(str(fps), 1, (100, 0, 0), (200, 200, 200))
            text_nums[fps] = text
        else:
            text = text_nums[fps]
        tx, ty = text.get_size()
        sx, sy = self.screen.get_size()
        self.screen.fill((0, 0, 0), (sx-tx, sy-ty, tx, ty))
        self.screen.blit(text, (sx-tx, sy-ty))

    def renderFlood(self):
        for i, row in enumerate(self.lightff):
            for j, num in enumerate(row):
                if num is None:
                    continue
                elif num not in text_nums:
                    text = modenine.render(str(num), 1, (255, 0, 0))
                    text_nums[num] = text
                else:
                    text = text_nums[num]
                self.screen.blit(text, (i*16, j*16))

    def animate(self):
        self.progress.stopFixingLight()

        for ghost in self.ghosts:
            ghost.animate(self)
        if self.player.animate(self):
            return True
        redraw_lights = False
        for blob in self.blobs[:]:
            redraw_lights = redraw_lights or blob.animate(self)

        # how many players does it take to fix a light?
        if self.progress:
            redraw_lights = redraw_lights or self.progress.animate(self)

        # light changed - must redraw
        if redraw_lights:
            self.renderLights()
            self.renderPlayArea()

        for ghost in self.ghosts:
            # check for player demise
            if ghost.testCollision(self.player):
                return True

            # zapping ghosts?
            if ghost.zapping:
                sounds['zap%s'%random.randint(1,4)].play()
                self.blobs.append(Blob(ghost.position, ghost.zapping))
                x, y = ghost.zapping
                ghost.recharging = time.time() + ghost.RECHARGE_TIME

        return False

    def events(self):
        event = pygame.event.poll()
        while event.type != NOEVENT:
            if event.type == NOEVENT:
                pass
            elif event.type == QUIT:
                return True
            elif event.type == KEYDOWN:
                if self.keyDown(event.key, event.mod):
                    return True
            elif event.type == KEYUP:
                if self.keyUp(event.key):
                    return True
            elif event.type == WALK_SOUND_EVENT:
                self.player.walk_channel = None
            elif event.type == REPAIR_SOUND_EVENT:
                self.player.repair_channel = None
            event = pygame.event.poll()
        return False

    def keyUp(self, key):
        if key == K_LEFT:
            self.player.stopMoveLeft()
        elif key == K_RIGHT:
            self.player.stopMoveRight()
        elif key == K_UP:
            self.player.stopMoveUp()
        elif key == K_DOWN:
            self.player.stopMoveDown()

    def keyDown(self, key, mod):
        if key == K_p:
            self.pause = not self.pause
        elif key == K_f:
            self.show_fps = not self.show_fps
            self.render()
        elif key == K_s and mod & KMOD_SHIFT:
            if self.step: self.step = 0
            else: self.step = 1
        elif key == K_s:
            self.step = 1
        elif key == K_l:
            self.renderFlood()
            pygame.display.flip()
        elif key == K_e:
            self.expose = not self.expose
            self.render()
            pygame.display.flip()
        elif key == K_ESCAPE:
            return True
        elif key == K_LEFT:
            self.player.startMoveLeft()
        elif key == K_RIGHT:
            self.player.startMoveRight()
        elif key == K_UP:
            self.player.startMoveUp()
        elif key == K_DOWN:
            self.player.startMoveDown()
        return False

def calculateDistance(a, b):
    ax, ay = a; bx, by = b
    return math.sqrt((bx-ax)**2 + (by-ay)**2)

if __name__ == '__main__':
    try:
        win = Game().main()
        if win: print 'you win!'
        else: print 'you lose!'
    except Quit:
        pass

